import torch
import time
    
def worker_init_fn_inference(worker_id):
    worker_info = torch.utils.data.get_worker_info()
    dataset = worker_info.dataset
    total_workers = worker_info.num_workers
    current_worker = worker_id
    
    pos_per_worker = len(dataset.files)//total_workers
    
    if(current_worker == total_workers -1): #Last worker
        dataset.files = dataset.files[pos_per_worker*current_worker:]
    else:
        dataset.files = dataset.files[pos_per_worker*current_worker: pos_per_worker*(current_worker+1)]
       
    assert(len(dataset.files)>0)



import argparse
from pathlib import Path
import pickle
import numpy as np
import torch
from torch.utils.data import DataLoader
import pandas as pd
import sys
import os
from tqdm import tqdm

from rnakinet.data_utils.dataloading import UnlimitedReadsInferenceDatasetFast5, UnlimitedReadsInferenceDatasetPod5
from rnakinet.models.model import RNAkinet
from rnakinet.models.model_experimental import RNAkinet_LastOnly

from rnakinet.data_utils.workers import worker_init_fn_inference

def run(args):
    print('CUDA', torch.cuda.is_available())
    reads_format = args.format
    files = list(Path(args.path).rglob(f'*.{reads_format}'))
    print(f'Number of {reads_format} files found: {len(files)}')
    if(len(files)==0):
        raise Exception(f'No {reads_format} files found')
        
    kit_architecture = {
        'r9':RNAkinet, 
        'r10':RNAkinet_LastOnly
    }
    base_dir = os.path.dirname(os.path.dirname(__file__))
    kit_checkpoint = {
        'r9':os.path.join(base_dir, 'models', 'rnakinet_r9.ckpt'), 
        'r10':os.path.join(base_dir, 'models', 'rnakinet_r10.ckpt')
    }
    print('Using checkpoint', kit_checkpoint[args.kit])
    
    model = kit_architecture[args.kit]()
    
    model.load_state_dict(torch.load(kit_checkpoint[args.kit], map_location='cpu')['state_dict'])
    model.eval()
    
    if torch.cuda.is_available() and not args.use_cpu:
        model.cuda()
    else:
        model.cpu()
        
    format_dataset_map = {
        'fast5': UnlimitedReadsInferenceDatasetFast5,
        'pod5': UnlimitedReadsInferenceDatasetPod5,
    }
    
    dset = format_dataset_map[reads_format](files=files, max_len=args.max_len, min_len=args.min_len, skip=args.skip)
    
    dataloader = DataLoader(
        dset, 
        batch_size=args.batch_size, 
        num_workers=min([args.max_workers, len(dset.files)]), 
        pin_memory=False, 
        worker_init_fn=worker_init_fn_inference
    )

    id_to_pred = {}
    with torch.no_grad():
        for batch in tqdm(dataloader):
            inputs, ids = batch
            if torch.cuda.is_available() and not args.use_cpu:
                inputs = inputs.cuda()
            else:
                inputs = inputs.cpu()
                
            outputs = model(inputs)
            
            readid_probs = zip(ids['readid'], outputs.cpu().numpy())
            for readid, probab in readid_probs:
                assert len(probab) == 1
                id_to_pred[readid] = probab[0]
                
    with open(args.output, 'wb') as handle:
        df = pd.DataFrame.from_dict(id_to_pred, orient='index').reset_index()
        df.columns = ['read_id', '5eu_mod_score']
        df['5eu_modified_prediction'] = df['5eu_mod_score'] > args.threshold
        df.to_csv(handle, index=False)
        
def main():
    
    parser = argparse.ArgumentParser(description='Run prediction on FAST5 files')
    parser.add_argument('--path', type=str, required=True, help='Path to the folder containing FAST5 files.')
    parser.add_argument('--kit', type=str, default='r9', help='Sequencing kit used to produce FAST5 files. Model will be selected based on this.', choices=['r9','r10'])
    parser.add_argument('--output', type=str, required=True, help='Path to the output csv file for pooled predictions.')
    parser.add_argument('--max-workers', type=int, default=16, help='Maximum number of workers for data loading')
    parser.add_argument('--batch-size', type=int, default=1, help='Batch size for data loading')
    parser.add_argument('--max-len', type=int, default=400000, help='Maximum length of the signal sequence to process')
    parser.add_argument('--min-len', type=int, default=5000, help='Minimum length of the signal sequence to process')
    parser.add_argument('--skip', type=int, default=5000, help='How many signal steps to skip at the beginning of each sequence (trimming)')
    parser.add_argument('--threshold', type=float, default=0.5, help='Threshold for the predictions to be considered positives')
    parser.add_argument('--use-cpu', action='store_true', help='Use CPU for computation instead of GPU')
    parser.add_argument('--format', type=str, default='fast5', help='Format of read files', choices=['fast5','pod5'])

    
    args = parser.parse_args(sys.argv[1:])
    run(args)

if __name__ == "__main__":
    main()


import argparse
import pandas as pd
import pysam
import pickle
import numpy as np
import sys

def run(args):
    bamfile = pysam.AlignmentFile(args.transcriptome_bam, 'rb')
    read_to_transcript_data = [
        {'read_id':read.query_name, 'transcript':read.reference_name} 
        for read in bamfile 
        if (not read.is_unmapped and not read.is_supplementary)
    ]
    bamfile.close()
    read_to_transcript = pd.DataFrame(read_to_transcript_data)
    
    read_to_score = pd.read_csv(args.predictions)

    read_transcript_score = read_to_score.merge(read_to_transcript, how='left', on='read_id')
    read_transcript_score = read_transcript_score.dropna()
    transcript_groups = read_transcript_score.groupby('transcript')
    transcript_agg = transcript_groups['5eu_modified_prediction'].mean().reset_index()
    transcript_agg = transcript_agg.rename(columns={'5eu_modified_prediction':'percentage_modified'})
    transcript_agg['reads'] = transcript_groups.size().values
    assert len(transcript_agg)>0, 'Empty transcript level prediction file, check if read ids in predictions also appear in the bam file'
    
    transcript_agg = add_predicted_halflifes(transcript_agg, tl=args.tl)
    
    transcript_agg.to_csv(args.output)
    
def add_predicted_halflifes(df, tl):
    col = 'percentage_modified'
    df['pred_t5'] = -tl * np.log(2) / np.log(1-df[col]) 
    df = df.replace([np.inf, -np.inf], np.nan)
    df = df[~df['pred_t5'].isna()]
    return df
    
def main():
    parser = argparse.ArgumentParser(description="Create a table of genes and their modification predictions stats")
    parser.add_argument("--transcriptome-bam", help="transcriptome bam file path")
    parser.add_argument("--predictions", help="predictions for individual reads")
    parser.add_argument('--tl', type=float, required=True, help='Time parameter for the decay equation')
    parser.add_argument("--output", help="path to the transcript output table")
    
    args = parser.parse_args(sys.argv[1:])
    run(args)
    


import torch

class Permute(torch.nn.Module):
    def forward(self, x):
        return x.permute(0, 2, 1)


class RNNPooling(torch.nn.Module):
    def forward(self, x):
        output = x[0]  # All states
        h_n = x[1]  # Last hidden states
        max_pooled, _ = torch.max(output, dim=1)
        mean_pooled = torch.mean(output, dim=1)
        last_states = h_n.permute(1, 0, 2).flatten(start_dim=1)
        concat = torch.cat(
            [max_pooled, mean_pooled, last_states], dim=1)
        return concat


class RNAkinet(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.head = torch.nn.Sequential(
            torch.nn.Conv1d(in_channels=1, out_channels=8, kernel_size=3),
            torch.nn.ReLU(),
            torch.nn.MaxPool1d(kernel_size=3),
            torch.nn.Conv1d(in_channels=8, out_channels=16, kernel_size=3),
            torch.nn.ReLU(),
            torch.nn.MaxPool1d(kernel_size=3),
            torch.nn.Conv1d(in_channels=16, out_channels=32, kernel_size=3),
            torch.nn.ReLU(),
            torch.nn.MaxPool1d(kernel_size=3),
            torch.nn.Conv1d(in_channels=32, out_channels=64, kernel_size=3),
            torch.nn.ReLU(),
            torch.nn.MaxPool1d(kernel_size=3),
            torch.nn.Conv1d(in_channels=64, out_channels=128, kernel_size=3),
            torch.nn.ReLU(),
            torch.nn.MaxPool1d(kernel_size=3),
            
            
            Permute(),
            torch.nn.GRU(input_size=128, hidden_size=32, num_layers=1,
                         batch_first=True, bidirectional=True, dropout=0.0),
            RNNPooling(),
            torch.nn.Linear(1*(2*3*32), 30),
            torch.nn.ReLU(),
            torch.nn.Linear(30, 1),

        )

    def forward(self, x):
        out = self.head(x)
        out = torch.sigmoid(out)
        return out
